#!/usr/bin/env python3
# Copyright (c) 2004-present Facebook All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

from enum import Enum
from functools import partial
from typing import Dict, Type
from dataclasses_json import config


class MissingEnumException(Exception):
    def __init__(self, variable: Enum) -> None:
        self.enum_type = str(type(variable))

    def __str__(self) -> str:
        return f"Try to encode missing value of enum {self.enum_type}"


def enum_field_metadata(enum_type: Type[Enum]) -> Dict[str, dict]:
    def encode_enum(value: Enum) -> str:
        if value.value == "":
            raise MissingEnumException(value)
        return value.value

    def decode_enum(t: Type[Enum], value: str) -> Enum:
        return t(value)

    return config(
        encoder=encode_enum,
        decoder=partial(decode_enum, enum_type),
    )



#!/usr/bin/env python3
# Copyright (c) 2004-present Facebook All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

ENUM_DIRNAME = "enum"
INPUT_DIRNAME = "input"

SPACES = " " * 4



#!/usr/bin/env python3
# Copyright (c) 2004-present Facebook All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import os
import re

from .constant import SPACES


def camel_case_to_lower_case(string: str) -> str:
    return re.sub(r"(?<!^)(?=[A-Z])", "_", string).lower()


class CodeChunk:
    class Block:
        def __init__(self, codegen: "CodeChunk"):
            self.gen = codegen

        def __enter__(self):
            self.gen.indent()
            return self.gen

        def __exit__(self, exc_type, exc_val, exc_tb):
            self.gen.unindent()

    def __init__(self):
        self.lines = []
        self.level = 0

    def indent(self):
        self.level += 1

    def unindent(self):
        if self.level > 0:
            self.level -= 1

    @property
    def indent_string(self):
        return self.level * SPACES

    def write(self, value: str, *args, **kwargs):
        if value != "":
            value = self.indent_string + value
        if args or kwargs:
            value = value.format(*args, **kwargs)

        self.lines.append(value)

    def write_lines(self, lines):
        for line in lines:
            self.lines.append(self.indent_string + line)

    def block(self):
        return self.Block(self)

    def write_block(self, block_header: str, *args, **kwargs):
        self.write(block_header, *args, **kwargs)
        return self.block()

    def __str__(self):
        return os.linesep.join(self.lines)



#!/usr/bin/env python3
# Copyright (c) 2004-present Facebook All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

from typing import Any, Dict, Union

from dataclasses_json import DataClassJsonMixin
from enum import Enum

from ..compiler.renderer_dataclasses import CustomScalar
from .enum_utils import MissingEnumException


def encode_value(
    value: Union[DataClassJsonMixin, Any, Enum],
    custom_scalars: Dict[str, CustomScalar] = {},
) -> Union[Dict[str, Any], str]:
    if isinstance(value, DataClassJsonMixin):
        return encode_variables(value.to_dict(), custom_scalars)
    elif isinstance(value, Enum):
        if value.value == "":
            raise MissingEnumException(value)
        return value.value
    for _, custom_scalar in custom_scalars.items():
        if custom_scalar.encoder and isinstance(value, custom_scalar.type):
            return custom_scalar.encoder(value)
    return value


def encode_variables(
    variables: Dict[str, Any], custom_scalars: Dict[str, CustomScalar] = {}
) -> Dict[str, Any]:
    new_variables: Dict[str, Any] = {}
    for key, value in variables.items():
        if isinstance(value, list):
            new_list = [encode_value(val, custom_scalars) for val in value]
            new_variables[key] = new_list
        else:
            new_variables[key] = encode_value(value, custom_scalars)
    return new_variables



#!/usr/bin/env python3
# Copyright (c) 2004-present Facebook All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

from dataclasses import dataclass, field
from typing import Any, List, Mapping, Union, cast

from graphql import (
    GraphQLEnumType,
    GraphQLInputObjectType,
    GraphQLList,
    GraphQLNonNull,
    GraphQLSchema,
    ListTypeNode,
    NonNullTypeNode,
    OperationDefinitionNode,
    TypeInfo,
    TypeInfoVisitor,
    TypeNode,
    Visitor,
    is_enum_type,
    is_scalar_type,
    parse,
    validate,
    visit,
)
from graphql.validation.rules.no_unused_fragments import NoUnusedFragmentsRule
from graphql.validation.specified_rules import specified_rules


@dataclass
class ParsedField:
    name: str
    type: str
    is_list: bool
    nullable: bool
    default_value: Any = None


@dataclass
class ParsedEnum:
    name: str
    values: Mapping[str, Any]


@dataclass
class ParsedObject:
    name: str
    fields: List[ParsedField] = field(default_factory=list)
    parents: List[str] = field(default_factory=list)
    children: List["ParsedObject"] = field(default_factory=list)
    inputs: List["ParsedObject"] = field(default_factory=list)
    input_enums: List[ParsedEnum] = field(default_factory=list)


@dataclass
class ParsedVariableDefinition:
    name: str
    type: str
    nullable: bool
    is_list: bool
    default_value: Any = None


@dataclass
class ParsedOperation:
    name: str
    type: str
    variables: List[ParsedVariableDefinition] = field(default_factory=list)
    children: List[ParsedObject] = field(default_factory=list)


NodeT = Union[ParsedOperation, ParsedObject]


@dataclass
class ParsedQuery:
    query: str
    objects: List[ParsedOperation] = field(default_factory=list)
    input_objects: List[ParsedObject] = field(default_factory=list)
    enums: List[ParsedEnum] = field(default_factory=list)
    internal_enums: List[ParsedEnum] = field(default_factory=list)
    internal_inputs: List[ParsedObject] = field(default_factory=list)
    fragment_objects: List[ParsedObject] = field(default_factory=list)
    used_fragments: List[str] = field(default_factory=list)


class FieldToTypeMatcherVisitor(Visitor):
    def __init__(self, schema: GraphQLSchema, type_info: TypeInfo, query: str):
        self.schema = schema
        self.type_info = type_info
        self.query = query
        self.parsed = ParsedQuery(query=self.query)
        self.dfs_path: List[NodeT] = []

    def push(self, obj: NodeT):
        self.dfs_path.append(obj)

    def pull(self) -> NodeT:
        return self.dfs_path.pop()

    @property
    def current(self) -> NodeT:
        return self.dfs_path[-1]

    # Document
    def enter_operation_definition(self, node: OperationDefinitionNode, *_args):
        name, operation = node.name, node.operation

        variables = []

        input_objects = []
        for var in node.variable_definitions:
            ptype, nullable, is_list, var_type = self.__variable_type_to_python(
                var.type
            )
            default_value = var.default_value
            variables.append(
                ParsedVariableDefinition(
                    name=var.variable.name.value,
                    type=ptype,
                    nullable=nullable,
                    is_list=is_list,
                    default_value=default_value.value  # type: ignore
                    if default_value
                    else None,
                )
            )

            var_name = var_type.name.value

            if isinstance(self.schema.type_map[var_name], GraphQLEnumType):
                enum_type = cast(GraphQLEnumType, self.schema.type_map[var_name])
                enum_name = enum_type.name
                if not any(e.name == enum_name for e in self.parsed.enums):
                    parsed_enum = ParsedEnum(
                        name=enum_type.name,
                        values={
                            val_name: val_value.value or val_name
                            for val_name, val_value in enum_type.values.items()
                        },
                    )
                    self.parsed.enums.append(parsed_enum)
            elif isinstance(self.schema.type_map[var_name], GraphQLInputObjectType):
                input_type = cast(
                    GraphQLInputObjectType, self.schema.type_map[var_name]
                )
                obj = self.parse_input_object(input_type)
                input_objects.append(obj)

        for obj in input_objects:
            self.parsed.input_objects.append(obj)

        name_value = name.value if name is not None else ""
        parsed_op = ParsedOperation(
            name=name_value,
            type=str(operation.value),
            variables=variables,
            children=[ParsedObject(name=f"{name_value}Data")],
        )

        self.parsed.objects.append(parsed_op)  # pylint:disable=no-member
        self.push(parsed_op)
        self.push(parsed_op.children[0])  # pylint:disable=unsubscriptable-object

        return node

    def parse_input_object(self, obj_type):
        obj = ParsedObject(name=str(obj_type))
        for field_name, field_value in obj_type.fields.items():
            field, field_obj_type, parsed_enum = self.__parse_field(
                field_name, field_value.type
            )
            if parsed_enum is not None:
                obj.input_enums.append(parsed_enum)
                self.parsed.internal_enums.append(
                    parsed_enum
                )  # pylint:disable=no-member
            obj.fields.append(field)
            if field_obj_type is not None:
                field_obj = self.parse_input_object(field_obj_type)
                obj.inputs.append(field_obj)
                self.parsed.internal_inputs.append(field_obj)
        return obj

    # def enter_selection_set(self, node, *_):
    #     return node

    def leave_selection_set(self, node, *_):
        self.pull()
        return node

    # Fragments

    def enter_fragment_definition(self, node, *_):
        # Same as operation definition
        obj = ParsedObject(name=node.name.value)
        self.parsed.fragment_objects.append(obj)  # pylint:disable=no-member
        self.push(obj)
        return node

    def enter_fragment_spread(self, node, *_):
        self.current.parents.append(node.name.value)
        self.parsed.used_fragments.append(node.name.value)
        return node

    # def enter_inline_fragment(self, node, *_):
    #     return node
    #
    # def leave_inline_fragment(self, node, *_):
    #     return node

    # Field

    def enter_field(self, node, *_):
        name = node.alias.value if node.alias else node.name.value
        graphql_type = self.type_info.get_type()

        field, obj_type, parsed_enum = self.__parse_field(name, graphql_type)
        if parsed_enum is not None:
            self.parsed.enums.append(parsed_enum)  # pylint:disable=no-member
        self.current.fields.append(field)

        if obj_type is not None:
            obj = ParsedObject(name=str(obj_type))
            self.current.children.append(obj)
            self.push(obj)

        return node

    def __parse_field(self, name, graphql_type):
        (
            type_name,
            is_list,
            nullable,
            underlying_graphql_type,
        ) = self.__scalar_type_to_python(graphql_type)

        parsed_field = ParsedField(
            name=name, type=type_name, is_list=is_list, nullable=nullable
        )
        parsed_enum = None

        if not is_scalar_type(underlying_graphql_type):
            if is_enum_type(underlying_graphql_type):
                enum_type = cast(
                    GraphQLEnumType, self.schema.type_map[underlying_graphql_type.name]
                )
                name = enum_type.name
                if not any(e.name == name for e in self.parsed.enums):
                    parsed_enum = ParsedEnum(
                        name=enum_type.name,
                        values={
                            name: value.value or name
                            for name, value in enum_type.values.items()
                        },
                    )
            else:
                return parsed_field, underlying_graphql_type, parsed_enum

        return parsed_field, None, parsed_enum

    @staticmethod
    def __scalar_type_to_python(scalar):
        nullable = True
        is_list = False
        if isinstance(scalar, GraphQLNonNull):
            nullable = False
            scalar = scalar.of_type

        if isinstance(scalar, GraphQLList):
            scalar = scalar.of_type
            if isinstance(scalar, GraphQLNonNull):
                scalar = scalar.of_type
                nullable = False
            is_list = True

        return scalar.name, is_list, nullable, scalar

    @staticmethod
    def __variable_type_to_python(var_type: TypeNode):
        nullable = True
        is_list = False
        if isinstance(var_type, NonNullTypeNode):
            nullable = False
            var_type = var_type.type
        if isinstance(var_type, ListTypeNode):
            is_list = True
            var_type = var_type.type
        if isinstance(var_type, NonNullTypeNode):
            nullable = False
            var_type = var_type.type

        return var_type.name.value, nullable, is_list, var_type  # type: ignore


class InvalidQueryError(Exception):
    def __init__(self, errors):
        self.errors = errors
        message = "\n".join(str(err) for err in errors)
        super().__init__(message)


class QueryParser:
    def __init__(self, schema: GraphQLSchema):
        self.schema = schema
        self.__jinja2_env = None

    def parse(
        self,
        query: str,
        full_fragments: str = "",
        should_validate: bool = True,
    ) -> ParsedQuery:
        query_document_ast = parse("".join([full_fragments, query]))
        document_ast = parse(query)

        if should_validate:
            errors = validate(
                self.schema,
                query_document_ast,
                [rule for rule in specified_rules if rule is not NoUnusedFragmentsRule],
            )
            if errors:
                raise InvalidQueryError(errors)

        type_info = TypeInfo(self.schema)
        visitor = FieldToTypeMatcherVisitor(self.schema, type_info, query)
        visit(document_ast, TypeInfoVisitor(type_info, visitor))
        result = visitor.parsed
        return result



#!/usr/bin/env python3
# Copyright (c) 2004-present Facebook All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

from typing import Dict, List, Optional, Type, Any, Callable

import importlib.util

from graphql import GraphQLSchema

from .query_parser import (
    ParsedField,
    ParsedObject,
    ParsedOperation,
    ParsedQuery,
    ParsedVariableDefinition,
)
from .utils_codegen import (
    CodeChunk,
    camel_case_to_lower_case,
)

from importlib.abc import Loader

from marshmallow.fields import Field as MarshmallowField

from dataclasses import dataclass

DEFAULT_MAPPING = {
    "ID": "str",
    "String": "str",
    "Int": "int",
    "Float": "Number",
    "Boolean": "bool",
}


@dataclass
class CustomScalar:
    name: str
    type: Type[Any]
    encoder: Optional[Callable[..., Any]] = None
    decoder: Optional[Callable[..., Any]] = None
    mm_field: Optional[MarshmallowField] = None


class DataclassesRenderer:
    def __init__(
        self, schema: GraphQLSchema, config_path: Optional[str] = None
    ) -> None:
        self.schema = schema
        self.config_path = config_path
        self.custom_scalars = {}
        if config_path is not None:
            spec = importlib.util.spec_from_file_location("config", config_path)
            config = importlib.util.module_from_spec(spec)
            assert isinstance(spec.loader, Loader)
            spec.loader.exec_module(config)
            assert hasattr(config, "custom_scalars"), "Custom scalars is not in config"
            self.custom_scalars = getattr(config, "custom_scalars")

    def render(
        self,
        parsed_query: ParsedQuery,
        fragment_name_to_importpath: Dict[str, str],
        enum_name_to_importpath: Dict[str, str],
        input_name_to_importpath: Dict[str, str],
        config_importpath: Optional[str],
    ) -> str:
        buffer = CodeChunk()
        self.__write_file_header(buffer)
        buffer.write("from dataclasses import dataclass, field")
        self.__render_customer_scalars_imports(buffer, config_importpath)
        buffer.write("from gql_client.runtime.variables import encode_variables")
        buffer.write("from gql import gql, Client")
        buffer.write("from gql.transport.exceptions import TransportQueryError")
        buffer.write("from functools import partial")
        buffer.write("from numbers import Number")
        buffer.write(
            "from typing import Any, AsyncGenerator, Dict, List, Generator, Optional"
        )
        buffer.write("from time import perf_counter")
        buffer.write("from dataclasses_json import DataClassJsonMixin, config")
        buffer.write("")
        for fragment_name in sorted(set(parsed_query.used_fragments)):
            importpath = fragment_name_to_importpath[fragment_name]
            buffer.write(
                f"from {importpath} import {fragment_name}, QUERY as {fragment_name}Query"
            )
            buffer.write("")
        enum_names = set()
        for enum in parsed_query.enums:
            enum_names.add(enum.name)
        if enum_names:
            buffer.write(
                "from gql_client.runtime.enum_utils import enum_field_metadata"
            )
            for enum_name in sorted(enum_names):
                importpath = enum_name_to_importpath[enum_name]
                buffer.write(f"from {importpath} import {enum_name}")
            buffer.write("")
        input_object_names = set()
        for input_object in parsed_query.input_objects:
            input_object_names.add(input_object.name)
        if input_object_names:
            for input_object_name in sorted(input_object_names):
                importpath = input_name_to_importpath[input_object_name]
                buffer.write(f"from {importpath} import {input_object_name}")
            buffer.write("")

        sorted_objects = sorted(
            parsed_query.objects,
            key=lambda obj: 1 if isinstance(obj, ParsedOperation) else 0,
        )
        for obj in sorted_objects:
            buffer.write("")
            if isinstance(obj, ParsedObject):
                self.__render_object(parsed_query, buffer, obj)
            elif isinstance(obj, ParsedOperation):
                self.__render_operation(parsed_query, buffer, obj, config_importpath)

        if parsed_query.fragment_objects:
            buffer.write("# fmt: off")
            if parsed_query.used_fragments:
                queries = [
                    f"{fragment_name}Query"
                    for fragment_name in sorted(set(parsed_query.used_fragments))
                ]
                buffer.write(f'QUERY: List[str] = {" + ".join(queries)} + ["""')
            else:
                buffer.write('QUERY: List[str] = ["""')
            buffer.write(parsed_query.query)
            buffer.write('"""]')
            buffer.write("")

        for fragment_obj in parsed_query.fragment_objects:
            self.__render_fragment(parsed_query, buffer, fragment_obj)

        return str(buffer)

    def render_enums(self, parsed_query: ParsedQuery) -> Dict[str, str]:
        result = {}

        for enum in parsed_query.enums + parsed_query.internal_enums:
            buffer = CodeChunk()
            self.__write_file_header(buffer)
            buffer.write("from enum import Enum")
            buffer.write("")
            buffer.write("")
            with buffer.write_block(f"class {enum.name}(Enum):"):
                for value_name, value in enum.values.items():
                    if isinstance(value, str):
                        value = f'"{value}"'

                    buffer.write(f"{value_name} = {value}")
                buffer.write('MISSING_ENUM = ""')
                buffer.write("")
                buffer.write("@classmethod")
                with buffer.write_block(
                    f'def _missing_(cls, value: object) -> "{enum.name}":'
                ):
                    buffer.write("return cls.MISSING_ENUM")
            buffer.write("")
            result[enum.name] = str(buffer)

        return result

    def render_input_objects(
        self, parsed_query: ParsedQuery, config_importpath: Optional[str]
    ) -> Dict[str, str]:
        result = {}

        for input_object in parsed_query.input_objects + parsed_query.internal_inputs:
            buffer = CodeChunk()
            self.__write_file_header(buffer)
            buffer.write("from dataclasses import dataclass, field")
            buffer.write("from functools import partial")
            self.__render_customer_scalars_imports(buffer, config_importpath)
            buffer.write("from numbers import Number")
            buffer.write(
                "from typing import Any, AsyncGenerator, Dict, List, Generator, Optional"
            )
            buffer.write("")
            buffer.write("from dataclasses_json import DataClassJsonMixin, config")
            buffer.write("")
            enum_names = set()
            for enum in input_object.input_enums:
                enum_names.add(enum.name)
            if enum_names:
                buffer.write(
                    "from gql_client.runtime.enum_utils import enum_field_metadata"
                )
                for enum_name in sorted(enum_names):
                    buffer.write(
                        f"from ..enum.{camel_case_to_lower_case(enum_name)} import {enum_name}"
                    )
                buffer.write("")
            input_object_names = set()
            for input_dep in input_object.inputs:
                input_object_names.add(input_dep.name)
            if input_object_names:
                for input_object_name in sorted(input_object_names):
                    buffer.write(
                        f"from ..input.{camel_case_to_lower_case(input_object_name)} "
                        f"import {input_object_name}"
                    )
                buffer.write("")

            buffer.write("")
            self.__render_object(parsed_query, buffer, input_object, True)
            result[input_object.name] = str(buffer)

        return result

    def __render_object(
        self,
        parsed_query: ParsedQuery,
        buffer: CodeChunk,
        obj: ParsedObject,
        is_input: bool = False,
    ) -> None:
        class_parents = (
            "(DataClassJsonMixin)" if not obj.parents else f'({", ".join(obj.parents)})'
        )

        buffer.write("@dataclass(frozen=True)")
        with buffer.write_block(f"class {obj.name}{class_parents}:"):
            # render child objects
            children_names = set()
            for child_object in obj.children:
                if child_object.name not in children_names:
                    self.__render_object(parsed_query, buffer, child_object, is_input)
                children_names.add(child_object.name)

            # render fields
            fields = obj.fields
            if is_input:
                fields = self.__sort_fields(parsed_query, obj.fields)

            for field in fields:
                self.__render_field(parsed_query, buffer, field, is_input)

            # pass if not children or fields
            if not (obj.children or obj.fields):
                buffer.write("pass")

        buffer.write("")

    def __render_fragment(
        self, parsed_query: ParsedQuery, buffer: CodeChunk, obj: ParsedObject
    ) -> None:
        class_parents = (
            "(DataClassJsonMixin)" if not obj.parents else f'({", ".join(obj.parents)})'
        )

        buffer.write("@dataclass(frozen=True)")
        with buffer.write_block(f"class {obj.name}{class_parents}:"):

            # render child objects
            children_names = set()
            for child_object in obj.children:
                if child_object.name not in children_names:
                    self.__render_object(parsed_query, buffer, child_object)
                children_names.add(child_object.name)

            # render fields
            for field in obj.fields:
                self.__render_field(parsed_query, buffer, field)

        buffer.write("")

    def __render_operation(
        self,
        parsed_query: ParsedQuery,
        buffer: CodeChunk,
        parsed_op: ParsedOperation,
        config_importpath: Optional[str],
    ) -> None:
        buffer.write("# fmt: off")
        if len(parsed_query.used_fragments):
            queries = [
                f"{fragment_name}Query"
                for fragment_name in sorted(set(parsed_query.used_fragments))
            ]
            buffer.write(f'QUERY: List[str] = {" + ".join(queries)} + ["""')
        else:
            buffer.write('QUERY: List[str] = ["""')
        buffer.write(parsed_query.query)
        buffer.write('"""')
        buffer.write("]")
        buffer.write("")
        buffer.write("")
        with buffer.write_block(f"class {parsed_op.name}:"):
            # Render children
            for child_object in parsed_op.children:
                self.__render_object(parsed_query, buffer, child_object)

            # Execution functions
            if parsed_op.variables:
                vars_args = ", " + ", ".join(
                    [
                        self.__render_variable_definition(var)
                        for var in parsed_op.variables
                    ]
                )
                variables_dict = (
                    "{"
                    + ", ".join(
                        f'"{var.name}": {var.name}' for var in parsed_op.variables
                    )
                    + "}"
                )
            else:
                vars_args = ""
                variables_dict = "{}"

            assert len(parsed_op.children) == 1
            child = parsed_op.children[0]
            assert len(child.fields) == 1
            query = child.fields[0]
            query_name = query.name
            query_result_type = f"{query.type}"
            if (
                query_result_type not in DEFAULT_MAPPING.keys()
                and query_result_type not in self.custom_scalars.keys()
            ):
                query_result_type = f"{parsed_op.name}Data.{query.type}"
            if query.nullable:
                query_result_type = f"Optional[{query_result_type}]"
            if query.is_list:
                query_result_type = f"List[{query_result_type}]"
            if parsed_op.type in ["query", "mutation"]:
                self.__write_execute_method(
                    buffer,
                    vars_args,
                    query_result_type,
                    variables_dict,
                    parsed_op.name,
                    query_name,
                    config_importpath,
                )
                self.__write_async_execute_method(
                    buffer,
                    vars_args,
                    query_result_type,
                    variables_dict,
                    parsed_op.name,
                    query_name,
                    config_importpath,
                )
            else:
                self.__write_subscribe_method(
                    buffer,
                    vars_args,
                    query_result_type,
                    variables_dict,
                    parsed_op.name,
                    query_name,
                    config_importpath,
                )
                self.__write_async_subscribe_method(
                    buffer,
                    vars_args,
                    query_result_type,
                    variables_dict,
                    parsed_op.name,
                    query_name,
                    config_importpath,
                )

    def __render_customer_scalars_imports(
        self, buffer: CodeChunk, config_importpath: Optional[str]
    ) -> None:
        if not config_importpath:
            return
        scalar_types = set()
        for _, custom_scalar in self.custom_scalars.items():
            if custom_scalar.type.__module__ != "builtins":
                scalar_types.add(custom_scalar.type.__name__)
        types_import_line = ", " + ", ".join(scalar_types) if scalar_types else ""
        buffer.write(
            f"from {config_importpath} import custom_scalars{types_import_line}"
        )

    @staticmethod
    def __write_execute_method(
        buffer: CodeChunk,
        vars_args: str,
        query_result_type: str,
        variables_dict: str,
        operation_name: str,
        query_name: str,
        config_importpath: Optional[str],
    ) -> None:
        buffer.write("# fmt: off")
        buffer.write("@classmethod")
        with buffer.write_block(
            f"def execute(cls, client: Client{vars_args})" f" -> {query_result_type}:"
        ):
            buffer.write(f"variables: Dict[str, Any] = {variables_dict}")
            scalars = ", custom_scalars" if config_importpath else ""
            buffer.write(f"new_variables = encode_variables(variables{scalars})")
            buffer.write("response_text = client.execute(")
            buffer.write('    gql("".join(set(QUERY))), variable_values=new_variables')
            buffer.write(")")
            buffer.write(f"res = cls.{operation_name}Data.from_dict(response_text)")
            buffer.write(f"return res.{query_name}")
        buffer.write("")

    @staticmethod
    def __write_async_execute_method(
        buffer: CodeChunk,
        vars_args: str,
        query_result_type: str,
        variables_dict: str,
        operation_name: str,
        query_name: str,
        config_importpath: Optional[str],
    ) -> None:
        buffer.write("# fmt: off")
        buffer.write("@classmethod")
        with buffer.write_block(
            f"async def execute_async(cls, client: Client{vars_args})"
            f" -> {query_result_type}:"
        ):
            buffer.write(f"variables: Dict[str, Any] = {variables_dict}")
            scalars = ", custom_scalars" if config_importpath else ""
            buffer.write(f"new_variables = encode_variables(variables{scalars})")
            buffer.write("response_text = await client.execute_async(")
            buffer.write('    gql("".join(set(QUERY))), variable_values=new_variables')
            buffer.write(")")
            buffer.write(f"res = cls.{operation_name}Data.from_dict(response_text)")
            buffer.write(f"return res.{query_name}")
        buffer.write("")

    @staticmethod
    def __write_subscribe_method(
        buffer: CodeChunk,
        vars_args: str,
        query_result_type: str,
        variables_dict: str,
        operation_name: str,
        query_name: str,
        config_importpath: Optional[str],
    ) -> None:
        buffer.write("# fmt: off")
        buffer.write("@classmethod")
        with buffer.write_block(
            f"def subscribe(cls, client: Client{vars_args})"
            f" -> Generator[{query_result_type}, None, None]:"
        ):
            buffer.write(f"variables: Dict[str, Any] = {variables_dict}")
            scalars = ", custom_scalars" if config_importpath else ""
            buffer.write(f"new_variables = encode_variables(variables{scalars})")
            buffer.write("subscription = client.subscribe(")
            buffer.write('    gql("".join(set(QUERY))), variable_values=new_variables')
            buffer.write(")")
            with buffer.write_block("for response_text in subscription:"):
                buffer.write(f"res = cls.{operation_name}Data.from_dict(response_text)")
                buffer.write(f"yield res.{query_name}")
        buffer.write("")

    @staticmethod
    def __write_async_subscribe_method(
        buffer: CodeChunk,
        vars_args: str,
        query_result_type: str,
        variables_dict: str,
        operation_name: str,
        query_name: str,
        config_importpath: Optional[str],
    ) -> None:
        buffer.write("# fmt: off")
        buffer.write("@classmethod")
        with buffer.write_block(
            f"async def subscribe_async(cls, client: Client{vars_args})"
            f" -> AsyncGenerator[{query_result_type}, None]:"
        ):
            buffer.write(f"variables: Dict[str, Any] = {variables_dict}")
            scalars = ", custom_scalars" if config_importpath else ""
            buffer.write(f"new_variables = encode_variables(variables{scalars})")
            buffer.write("subscription = client.subscribe_async(")
            buffer.write('    gql("".join(set(QUERY))), variable_values=new_variables')
            buffer.write(")")
            with buffer.write_block("async for response_text in subscription:"):
                buffer.write(f"res = cls.{operation_name}Data.from_dict(response_text)")
                buffer.write(f"yield res.{query_name}")
        buffer.write("")

    @staticmethod
    def __sort_fields(
        parsed_query: ParsedQuery, fields: List[ParsedField]
    ) -> List[ParsedField]:
        def sort_key(field) -> int:
            if field.nullable:
                return 2
            return 0

        return sorted(fields, key=sort_key)

    def __render_variable_definition(self, var: ParsedVariableDefinition):
        var_type = DEFAULT_MAPPING.get(var.type, var.type)

        if var.type in self.custom_scalars.keys():
            var_type = self.custom_scalars[var.type].type.__name__

        if var.is_list:
            return f"{var.name}: List[{var_type}] = []"

        if not var.nullable:
            return f"{var.name}: {var_type}"

        return f'{var.name}: Optional[{var_type}] = {var.default_value or "None"}'

    def __render_field(
        self,
        parsed_query: ParsedQuery,
        buffer: CodeChunk,
        field: ParsedField,
        is_input: bool = False,
    ) -> None:
        enum_names = [e.name for e in parsed_query.enums + parsed_query.internal_enums]
        is_enum = field.type in enum_names
        suffix = ""
        field_type = DEFAULT_MAPPING.get(field.type, field.type)
        if field.is_list:
            field_type = f"List[{field_type}]"

        if is_enum:
            suffix = f" = field(metadata=enum_field_metadata({field_type}))"

        if field_type in self.custom_scalars.keys():
            if (
                self.custom_scalars[field_type].encoder
                or self.custom_scalars[field_type].decoder
                or self.custom_scalars[field_type].mm_field
            ):
                suffix = f' = field(metadata=config(encoder=custom_scalars["{field_type}"].encoder, decoder=custom_scalars["{field_type}"].decoder, mm_field=custom_scalars["{field_type}"].mm_field))'
            field_type = self.custom_scalars[field_type].type.__name__

        if field.nullable:
            if is_input:
                suffix = f" = {field.default_value}"
            buffer.write(f"{field.name}: Optional[{field_type}]{suffix}")
        else:
            buffer.write(f"{field.name}: {field_type}{suffix}")

    @staticmethod
    def __write_file_header(buffer: CodeChunk) -> None:
        buffer.write("#!/usr/bin/env python3")
        buffer.write("# @" + "generated AUTOGENERATED file. Do not Change!")
        buffer.write("")



#!/usr/bin/env python3
# Copyright (c) 2004-present Facebook All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import glob
import os
from typing import Dict, Tuple, List, Optional

from graphql import GraphQLSchema
from graphql.language.parser import parse
from graphql.utilities.find_deprecated_usages import find_deprecated_usages

from .constant import ENUM_DIRNAME, INPUT_DIRNAME
from .query_parser import InvalidQueryError, QueryParser
from .renderer_dataclasses import DataclassesRenderer
from .utils_codegen import CodeChunk, camel_case_to_lower_case
from .utils_schema import get_query_details, QueryType


def assert_rendered_file(file_name: str, file_content: str, rendered: str) -> None:
    assert (
        rendered == file_content
    ), f"""Generated file name {file_name} does
            not match compilation result:
            exising file:
            {file_content}
            compilation result:
            {rendered}"""


def safe_remove(fname: str) -> None:
    try:
        os.remove(fname)
    except BaseException:
        pass


def add_init_file(pkg_name: str) -> None:
    init_file_path = os.path.join(pkg_name, "__init__.py")
    if not os.path.exists(init_file_path):
        with open(init_file_path, "w") as outfile:
            buffer = CodeChunk()
            buffer.write("#!/usr/bin/env python3")
            buffer.write("# Copyright (c) 2004-present Facebook All rights reserved.")
            buffer.write("# Use of this source code is governed by a BSD-style")
            buffer.write("# license that can be found in the LICENSE file.")
            buffer.write("")
            outfile.write(str(buffer))


def verify_or_write_rendered(filename: str, rendered: str, verify: bool) -> None:
    if verify:
        with open(filename, "r") as f:
            file_content = f.read()
            assert_rendered_file(filename, file_content, rendered)
    else:
        with open(filename, "w") as outfile:
            outfile.write(rendered)


def make_python_package(pkg_name: str) -> None:
    if not os.path.exists(pkg_name):
        os.makedirs(pkg_name)
    add_init_file(pkg_name)


def root_no_ext(filename):
    root, _s = os.path.splitext(filename)
    return root


def get_import_path(filename, target_dirname):
    importpath = (
        root_no_ext(os.path.relpath(filename, target_dirname))
        .replace("../../../", "....")
        .replace("../../", "...")
        .replace("../", "..")
        .replace("/", ".")
    )
    if not importpath.startswith("."):
        return "." + importpath
    return importpath


def process_file(
    filename: str,
    query: str,
    schema: GraphQLSchema,
    parser: QueryParser,
    renderer: DataclassesRenderer,
    fragment_library: Dict[str, Tuple[str, str]],
    enum_dir_path: str,
    input_dir_path: str,
    config_path: Optional[str],
    verify: bool = False,
) -> None:
    full_fragments = "".join(
        [
            query
            for fragment_filename, query in fragment_library.values()
            if fragment_filename != filename
        ]
    )

    target_filename = "".join([root_no_ext(filename), ".py"])
    target_dirname = os.path.dirname(filename)

    try:
        parsed = parser.parse(query, full_fragments)
        fragment_name_to_importpath = {
            name: get_import_path(details[0], target_dirname)
            for name, details in fragment_library.items()
        }

        enum_name_to_importpath = {}
        enums = renderer.render_enums(parsed)
        for enum_name, code in enums.items():
            target_enum_filename = os.path.join(
                enum_dir_path, "".join([camel_case_to_lower_case(enum_name), ".py"])
            )
            verify_or_write_rendered(target_enum_filename, code, verify)
            enum_name_to_importpath[enum_name] = get_import_path(
                target_enum_filename, target_dirname
            )

        input_name_to_importpath = {}
        input_objects = renderer.render_input_objects(
            parsed,
            get_import_path(config_path, input_dir_path) if config_path else None,
        )
        for input_object_name, code in input_objects.items():
            target_input_object_filename = os.path.join(
                input_dir_path,
                "".join([camel_case_to_lower_case(input_object_name), ".py"]),
            )
            verify_or_write_rendered(target_input_object_filename, code, verify)
            input_name_to_importpath[input_object_name] = get_import_path(
                target_input_object_filename, target_dirname
            )

        rendered = renderer.render(
            parsed,
            fragment_name_to_importpath,
            enum_name_to_importpath,
            input_name_to_importpath,
            get_import_path(config_path, target_dirname) if config_path else None,
        )
        verify_or_write_rendered(target_filename, rendered, verify)
    except (InvalidQueryError, AssertionError):
        if verify:
            print(f"Failed to verify graphql file {filename}")
        else:
            print(f"Failed to process graphql file {filename}")
        safe_remove(target_filename)
        raise


def run(
    schema: GraphQLSchema,
    graphql_library: str,
    verify: bool = False,
    allow_deprecated: bool = False,
    config_path: str = None,
) -> None:
    filenames = glob.glob(os.path.join(graphql_library, "**/*.graphql"), recursive=True)

    query_parser = QueryParser(schema)

    query_renderer = DataclassesRenderer(schema, config_path)

    enum_dir_path = os.path.join(graphql_library, ENUM_DIRNAME)
    input_dir_path = os.path.join(graphql_library, INPUT_DIRNAME)
    if not verify:
        py_filenames = glob.glob(
            os.path.join(graphql_library, "**/*.py"), recursive=True
        )
        for py_filename in py_filenames:
            os.unlink(py_filename)
        add_init_file(graphql_library)
        make_python_package(enum_dir_path)
        make_python_package(input_dir_path)

    operation_library: List[Tuple[str, str]] = []
    fragment_library: Dict[str, Tuple[str, str]] = {}

    for filename in filenames:
        with open(filename, "r") as fin:
            query = fin.read()
        parsed_query = parse(query)
        if not allow_deprecated:
            usages = find_deprecated_usages(schema, parsed_query)
            assert (
                len(usages) == 0
            ), f"Graphql file name {filename} uses deprecated fields {usages}"
        query_name, query_type = get_query_details(parsed_query)
        if query_type == QueryType.OPERATION:
            operation_library.append((filename, query))
        else:
            fragment_library[query_name] = (filename, query)
        dirname = os.path.dirname(filename)
        while os.path.normpath(dirname) != os.path.normpath(graphql_library):
            add_init_file(dirname)
            dirname = os.path.dirname(dirname)

    for filename, query in fragment_library.values():
        process_file(
            filename,
            query,
            schema,
            query_parser,
            query_renderer,
            fragment_library,
            enum_dir_path,
            input_dir_path,
            config_path,
            verify,
        )
    for filename, query in operation_library:
        process_file(
            filename,
            query,
            schema,
            query_parser,
            query_renderer,
            fragment_library,
            enum_dir_path,
            input_dir_path,
            config_path,
            verify,
        )



#!/usr/bin/env python3
# Copyright (c) 2004-present Facebook All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import setuptools
import os

this_directory = os.path.abspath(os.path.dirname(__file__))
long_description = ""
with open(os.path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setuptools.setup(
    name="py-gql-client",
    version="0.0.16",
    author="Facebook Inc.",
    description="Python GraphQL generated code client",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(include=["gql_client", "gql_client.*"]),
    license="BSD License",
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    python_requires=">=3.6",
    include_package_data=True,
    scripts=["bin/gql-compiler"],
    install_requires=[
        "gql>=v3.0.0a5",
        "dataclasses==0.6",
        "dataclasses-json==0.5.2",
    ],
)



#!/usr/bin/env python3
# Copyright (c) 2004-present Facebook All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import glob
from enum import Enum
import json
import os
import requests
from typing import Tuple

from graphql import (
    build_ast_schema,
    build_client_schema,
    get_introspection_query,
    GraphQLSchema,
)
from graphql.language.parser import parse
from graphql.language import OperationDefinitionNode, FragmentDefinitionNode
from graphql.language.ast import DocumentNode


class QueryType(Enum):
    OPERATION = 1
    FRAGMENT = 2


def load_introspection_from_server(url):
    query = get_introspection_query()
    request = requests.post(url, json={"query": query})
    if request.status_code == 200:
        return request.json()["data"]

    raise Exception(
        f"Request failure with {request.status_code} status code for query {query}"
    )


def load_introspection_from_file(filename):
    with open(filename, "r") as fin:
        return json.load(fin)


def load_schema(uri):
    introspection = (
        load_introspection_from_file(uri)
        if os.path.isfile(uri)
        else load_introspection_from_server(uri)
    )
    return build_client_schema(introspection)


def compile_schema_library(schema_library: str) -> GraphQLSchema:
    full_schema = ""
    # use the following line to use .graphqls files as well
    # os.path.join(schema_library, "**/*.graphql*"), recursive=True
    schema_filepaths = glob.glob(
        os.path.join(schema_library, "**/*.graphql"), recursive=True
    )
    for schema_filepath in schema_filepaths:
        with open(schema_filepath) as schema_file:
            full_schema = full_schema + schema_file.read()
    return build_ast_schema(parse(full_schema))


def get_query_details(document_ast: DocumentNode) -> Tuple[str, QueryType]:
    definitions = document_ast.definitions
    assert (
        len(definitions) == 1
    ), f"number of definitions in query is invalid: {len(definitions)}"
    definition = definitions[0]
    if isinstance(definition, OperationDefinitionNode):
        if not definition.name:
            raise AssertionError("operation has no name")
        return (definition.name.value, QueryType.OPERATION)
    elif not isinstance(definition, FragmentDefinitionNode):
        raise AssertionError("invalid definition")
    return (definition.name.value, QueryType.FRAGMENT)
